# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing


class TcpdumpParameters(typing.NamedTuple):
    capture_file: str
    interface: typing.Optional[str] = None
    capture_filter: typing.Optional[str] = None
    capture_timeout: typing.Optional[int] = None


def tcpdump_parameters(
        capture_file: str,
        interface: str = None,
        capture_filter: str = None,
        capture_timeout: int = None):
    """Get tcpdump parameters
    """
    return TcpdumpParameters(capture_file=capture_file,
                             interface=interface,
                             capture_filter=capture_filter,
                             capture_timeout=capture_timeout)
