# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import shiftstack


@shiftstack.skip_unless_has_shiftstack()
class ShiftstackNovaTest(testtools.TestCase):

    def test_list_shifstack_nodes(self):
        nodes = shiftstack.list_shiftstack_nodes()
        self.assertIsInstance(nodes, tobiko.Selection)
        self.assertNotEqual([], nodes)
        return nodes

    def test_find_shifstack_nodes(self):
        nodes = self.test_list_shifstack_nodes()
        node = shiftstack.find_shiftstack_node(name=nodes.first.name)
        self.assertEqual(nodes.first.id, node.id)
