# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


def get_file_size(file_name: str,
                  ssh_client: ssh.SSHClientType = None,
                  sudo: bool = None) \
        -> int:
    output = sh.execute(f'wc -c "{file_name}"',
                        ssh_client=ssh_client,
                        sudo=sudo).stdout.strip()
    size, _ = output.split(' ', 1)
    return int(size)


def assert_file_size(file_size: int,
                     file_name: str,
                     ssh_client: ssh.SSHClientType = None,
                     sudo: bool = None):
    size = get_file_size(file_name=file_name,
                         ssh_client=ssh_client,
                         sudo=sudo)
    tobiko.get_test_case().assertEqual(file_size, size)
